@@ -18,7 +18,7 @@ let (>>=) = Lwt.(>>=)
 
 module String_set = Set.Make(String)
 
-let version = "2.14.0"
+let version = "2.16.0"
 
 (* +-----------------------------------------------------------------+
    | History                                                         |
@@ -184,10 +184,12 @@ let collect_formatters buf pps f =
     raise exn
 
 let discard_formatters pps f =
+  match pps with [] -> f () | fpp :: pps_rest ->
   (* First flush all formatters. *)
   List.iter (fun pp -> Format.pp_print_flush pp ()) pps;
   (* Save all formatter functions. *)
-  let save = List.map (fun pp -> Format.pp_get_formatter_out_functions pp ()) pps in
+  let h = Format.pp_get_formatter_out_functions fpp () in
+  let save = h :: List.map (fun pp -> Format.pp_get_formatter_out_functions pp ()) pps_rest in
   let restore () =
     List.iter2
       (fun pp out_functions ->
@@ -197,9 +199,10 @@ let discard_formatters pps f =
   in
   (* Output functions. *)
   let out_functions = {
+      h with (* OCaml 5.4 added a new field *)
     Format.out_string = (fun _ _ _ -> ()); out_flush = ignore;
-    out_newline = ignore; out_spaces = ignore ; out_indent = ignore
-  } in
+    out_newline = ignore; out_spaces = ignore ; out_indent = ignore;
+  }[@warning "-23"] in
   (* Replace formatter functions. *)
   List.iter (fun pp -> Format.pp_set_formatter_out_functions pp out_functions) pps;
   try
@@ -305,7 +308,7 @@ let parse_default parse str eos_is_error =
     | exn ->
         Error ([], "Unknown parsing error (please report it to the utop project): " ^ Printexc.to_string exn)
 
-let parse_toplevel_phrase_default = parse_default Parse.toplevel_phrase
+let parse_toplevel_phrase_default = fun str -> parse_default !Toploop.parse_toplevel_phrase str
 let parse_toplevel_phrase = ref parse_toplevel_phrase_default
 
 let parse_use_file_default = parse_default Parse.use_file
@@ -350,8 +353,7 @@ let check_phrase phrase =
         (* Construct "let _ () = let module _ = struct <items> end in ()" in order to test
            the typing and compilation of [items] without evaluating them. *)
         let unit =
-          let (%.) a b = Longident.Ldot (a, b) in
-          with_loc loc (Lident "Stdlib" %. "Unit" %. "()")
+          with_loc loc (UTop_compat.longident_of_list ["Stdlib"; "Unit"; "()"])
         in
         let top_def =
           let open Ast_helper in
