@@ -23,24 +23,6 @@ let set_of_list = List.fold_left (fun set x -> String_set.add x set) String_set.
    | Utils                                                           |
    +-----------------------------------------------------------------+ *)
 
-(* Transform a non-empty list of strings into a long-identifier. *)
-let longident_of_list = function
-  | [] ->
-      invalid_arg "UTop_complete.longident_of_list"
-  | component :: rest ->
-      let rec loop acc = function
-        | [] -> acc
-        | component :: rest -> loop (Longident.Ldot(acc, component)) rest
-      in
-      loop (Longident.Lident component) rest
-
-(* Check whether an identifier is a valid one. *)
-let is_valid_identifier id =
-  id <> "" &&
-    (match id.[0] with
-       | 'A' .. 'Z' | 'a' .. 'z' |  '_' -> true
-       | _ -> false)
-
 let add id set = if is_valid_identifier id then String_set.add id set else set
 
 let lookup_env f x env =
@@ -387,14 +369,14 @@ let visible_modules () =
             Array.fold_left
               (fun acc fname ->
                 if Filename.check_suffix fname ".cmi" then
-                  String_set.add (String.capitalize_ascii (Filename.chop_suffix fname ".cmi")) acc
+                  String_set.add (lax_modname_from_cmi fname) acc
                 else
                   acc)
               acc
               (Sys.readdir (if dir = "" then Filename.current_dir_name else dir))
           with Sys_error _ ->
             acc)
-        String_set.empty @@ Load_path.get_paths ()
+        String_set.empty @@ UTop_compat.get_load_path ()
     )
 
 let field_name { ld_id = id } = Ident.name id
@@ -406,7 +388,11 @@ let add_fields_of_type decl acc =
         acc
     | Type_record (fields, _) ->
         List.fold_left (fun acc field -> add (field_name field) acc) acc fields
+#if OCAML_VERSION >= (5, 2, 0)
+    | Type_abstract _ ->
+#else 
     | Type_abstract ->
+#endif
         acc
     | Type_open ->
         acc
@@ -421,7 +407,11 @@ let add_names_of_type decl acc =
         List.fold_left (fun acc cstr -> add (constructor_name cstr) acc) acc constructors
     | Type_record (fields, _) ->
         List.fold_left (fun acc field -> add (field_name field) acc) acc fields
+#if OCAML_VERSION >= (5, 2, 0)
+    | Type_abstract _ ->
+#else 
     | Type_abstract ->
+#endif
         acc
     | Type_open ->
         acc
@@ -818,6 +808,7 @@ let complete ~phrase_terminator ~input =
       let prefix = String.sub input (loc.ofs1 + tlen) (String.length input - loc.ofs1 - tlen) in
       begin match Parse.longident (Lexing.from_string prefix) with
       | Longident.Ldot (lident, last_prefix) ->
+        let lident, last_prefix = destruct_ldot lident last_prefix in
         let set = names_of_module lident in
         let compls = lookup last_prefix (String_set.elements set) in
         let start = loc.idx1 + 1 + (String.length prefix - String.length last_prefix) in
@@ -839,7 +830,7 @@ let complete ~phrase_terminator ~input =
               (fun acc d -> add_files filter acc (Filename.concat d dir))
               String_map.empty
               (Filename.current_dir_name ::
-                (Load_path.get_paths ())
+                (UTop_compat.get_load_path ())
               )
 
           else
@@ -899,7 +890,7 @@ let complete ~phrase_terminator ~input =
               (fun acc d -> add_files filter acc (Filename.concat d dir))
               String_map.empty
               (Filename.current_dir_name ::
-                (Load_path.get_paths ())
+                (UTop_compat.get_load_path ())
               )
           else
             add_files filter String_map.empty (Filename.dirname file)
