@@ -53,7 +53,10 @@ let init_history () =
         return ()
     | Some fn ->
         Lwt.catch
-          (fun () -> LTerm_history.load UTop.history fn)
+          (fun () ->
+            let dn = Filename.dirname fn in
+            if not (Sys.file_exists dn) then Unix.mkdir dn 0o700;
+            LTerm_history.load UTop.history fn)
           (function
           | Unix.Unix_error (error, func, arg) ->
               Logs_lwt.err (fun m -> m "cannot load history from %S: %s: %s"
@@ -310,7 +313,7 @@ end = struct
     let comp = Ident.name id in
     match path with
     | None -> Longident.Lident comp
-    | Some path -> Longident.Ldot (path, comp)
+    | Some path -> ldot path comp
 
   let is_auto_printer_attribute (attr : Parsetree.attribute) =
     let name = attr.attr_name in
@@ -529,14 +532,14 @@ type rewrite_rule = {
   (* Whether the rule is enabled or not. *)
 }
 
-let longident_lwt_main_run = Longident.Ldot (Longident.Lident "Lwt_main", "run")
+let longident_lwt_main_run = longident_of_list ["Lwt_main"; "run"]
 let longident_async_thread_safe_block_on_async_exn =
-  Longident.(Ldot (Ldot (Lident "Async", "Thread_safe"), "block_on_async_exn"))
+  longident_of_list ["Async"; "Thread_safe"; "block_on_async_exn"]
 
 let rewrite_rules = [
   (* Rewrite Lwt.t expressions to Lwt_main.run <expr> *)
   {
-    type_to_rewrite = Longident.(Ldot (Lident "Lwt", "t"));
+    type_to_rewrite = longident_of_list ["Lwt"; "t"];
     path_to_rewrite = None;
     required_values = [longident_lwt_main_run];
     rewrite = (fun loc e ->
@@ -551,7 +554,7 @@ let rewrite_rules = [
   (* Rewrite Async.Defered.t expressions to
      Async.Thread_safe.block_on_async_exn (fun () -> <expr>). *)
   {
-    type_to_rewrite = Longident.(Ldot (Ldot (Lident "Async", "Deferred"), "t"));
+    type_to_rewrite = longident_of_list ["Async"; "Deferred"; "t"];
     path_to_rewrite = None;
     required_values = [longident_async_thread_safe_block_on_async_exn];
     rewrite = (fun loc e ->
@@ -744,10 +747,10 @@ let rec loop term =
   match phrase_opt with
     | Some phrase ->
         (* Rewrite toplevel expressions. *)
-        let count = S.value UTop_private.count in
         let phrase = rewrite phrase in
         let phrase =
           if UTop.get_create_implicits () then
+            let count = S.value UTop_private.count in
             let binding_name = Printf.sprintf "_%d" count in
             bind_expressions binding_name phrase
           else
@@ -968,6 +971,14 @@ module Emacs(M : sig end) = struct
   let process_checked_phrase phrase =
     (* Rewrite toplevel expressions. *)
     let phrase = rewrite phrase in
+    let phrase =
+      if UTop.get_create_implicits () then
+        let count = S.value UTop_private.count in
+        let binding_name = Printf.sprintf "_%d" count in
+        bind_expressions binding_name phrase
+      else
+        phrase
+    in
     try
       Env.reset_cache_toplevel ();
       ignore (execute_phrase true Format.std_formatter phrase);
@@ -1165,6 +1176,26 @@ end
    | Extra macros                                                    |
    +-----------------------------------------------------------------+ *)
 
+module Printtyp =
+#if OCAML_VERSION >= (5, 3, 0)
+  Out_type
+#else
+  Printtyp
+#endif
+
+#if OCAML_VERSION >= (5,4,0)
+  module Data_types = Data_types
+  let present_arg = function
+  | Typedtree.Arg x -> Some x
+  | Typedtree.Omitted () -> None
+#else
+  module Data_types = Types
+  let present_arg x = x
+#endif
+let lbl_res lbl = get_desc lbl.Data_types.lbl_res
+let cstr_res c = get_desc c.Data_types.cstr_res
+let cstr_tag c = c.Data_types.cstr_tag
+
 let typeof sid =
   let id = Parse.longident (Lexing.from_string sid) in
   let env = !Toploop.toplevel_env in
@@ -1187,7 +1218,7 @@ let typeof sid =
     with Not_found ->
     try
       let lbl_desc = Env.find_label_by_name id env in
-      let (path, ty_decl) = from_type_desc (get_desc lbl_desc.Types.lbl_res) in
+      let (path, ty_decl) = from_type_desc (lbl_res lbl_desc) in
       let id = Ident.create_local (Path.name path) in
       Some (Printtyp.tree_of_type_declaration id ty_decl Types.Trec_not)
     with Not_found ->
@@ -1203,9 +1234,9 @@ let typeof sid =
     with Not_found ->
     try
       let cstr_desc = Env.find_constructor_by_name id env in
-      match cstr_desc.Types.cstr_tag with
+      match cstr_tag cstr_desc with
       | _ ->
-        let (path, ty_decl) = from_type_desc (get_desc cstr_desc.Types.cstr_res) in
+        let (path, ty_decl) = from_type_desc (cstr_res cstr_desc) in
         let id = Ident.create_local (Path.name path) in
         Some (Printtyp.tree_of_type_declaration id ty_decl Types.Trec_not)
     with Not_found ->
@@ -1218,7 +1249,11 @@ let typeof sid =
   | Some osig ->
     let buf = Buffer.create 128 in
     let pp = Format.formatter_of_buffer buf in
+#if OCAML_VERSION >= (5, 3, 0)
+    Format.fprintf pp "%a" (Format_doc.compat !Toploop.print_out_signature) [osig];
+#else
     !Toploop.print_out_signature pp [osig];
+#endif
     Format.pp_print_newline pp ();
     let str = Buffer.contents buf in
     Lwt_main.run (Lazy.force LTerm.stdout >>= fun term -> render_out_phrase term str)
@@ -1443,6 +1478,7 @@ let load_inputrc () =
 let protocol_version = 1
 
 let main_aux ~initial_env =
+  Clflags.debug := true;
   Arg.parse args file_argument usage;
 #if OCAML_VERSION >= (5, 0, 0) && OCAML_VERSION < (5, 1, 0)
   Topcommon.load_topdirs_signature ();
@@ -1511,7 +1547,7 @@ exception Found of Env.t
 
 let get_required_label name args =
   match List.find (fun (lab, _) -> lab = Asttypes.Labelled name) args with
-  | _, x -> x
+  | _, x -> present_arg x
   | exception Not_found -> None
 
 let walk dir ~init ~f =
@@ -1584,6 +1620,10 @@ let () =
   Location.register_error_of_exn
     (function
       | Envaux.Error err ->
+#if OCAML_VERSION >= (5,3,0)
+        Some (Location.error_of_printer_file Envaux.report_error_doc err)
+#else
         Some (Location.error_of_printer_file Envaux.report_error err)
+#endif
       | _ -> None
     )
