@@ -2,23 +2,30 @@
 
 ;; Copyright: (c) 2011, Jeremie Dimino <jeremie@dimino.org>
 ;; Author: Jeremie Dimino <jeremie@dimino.org>
-;; URL: https://github.com/diml/utop
+;; URL: https://github.com/ocaml-community/utop
 ;; Licence: BSD3
 ;; Version: 1.11
-;; Package-Requires: ((emacs "24") (tuareg "2.2.0"))
+;; Package-Requires: ((emacs "26") (tuareg "2.2.0"))
 ;; Keywords: ocaml languages
 
 ;; This file is a part of utop.
 
 ;;; Commentary:
 
-;; See the README for more info;
-;; https://github.com/diml/utop
+;; This package provides interaction with an OCaml utop toplevel (REPL).
+;;
+;; utop.el has two components - a nice OCaml REPL with auto-completion and a
+;; minor mode (`utop-minor-mode'), which extends OCaml major modes
+;; (e.g. `caml-mode' and `tuareg-mode') with commands to evaluate forms directly
+;; in the REPL.
+;;
+;; See the "Integration with Emacs" section of the README for more info.
 
 ;;; Code:
 
 (require 'easymenu)
 (require 'pcase)
+(require 'seq)
 (require 'tabulated-list)
 (require 'tuareg)
 
@@ -57,60 +64,65 @@ SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.")
 ;; +-----------------------------------------------------------------+
 
 (defgroup utop nil
-  "A toplevel for the ocaml programming language which interact
+  "A toplevel for the ocaml programming language which interacts
 with Emacs to provide an enhanced environment."
-  :tag "The Caml Emacs-Lisp Toplevel"
-  :version "1.0"
-  :group 'applications)
+  :group 'applications
+  :link '(url-link :tag "GitHub" "https://github.com/ocaml-community/utop")
+  :link '(emacs-commentary-link :tag "Commentary" "utop"))
 
 (defcustom utop-command "utop -emacs"
   "The command to execute for utop."
   :type 'string
-  :group 'utop)
+  :safe 'stringp)
 
 (defcustom utop-edit-command t
   "Whether to read the command from the minibuffer before running utop.
 
 If nil, `utop-command' will be used without modification."
   :type 'boolean
-  :group 'utop)
+  :safe 'booleanp)
 
-(defcustom utop-prompt 'utop-default-prompt
+(defcustom utop-prompt #'utop-default-prompt
   "The function which create the prompt for utop."
   :type 'function
-  :group 'utop)
+  :safe 'functionp)
 
 (defcustom utop-mode-hook nil
   "A hook that gets run when `utop-mode' is entered."
-  :type 'hook
-  :group 'utop)
+  :type 'hook)
 
 (defcustom utop-exit-hook nil
   "A hook that is run whenever `utop' is exited.
 This hook is only run if exiting actually kills the buffer."
-  :type 'hook
-  :group 'utop)
+  :type 'hook)
 
 (defcustom utop-load-packages-without-asking nil
   "Load packages from file local variables without asking"
   :type 'boolean
-  :group 'utop)
+  :safe 'booleanp)
+
+(defcustom utop-capf-wait-interval 0.01
+  "Length of time to wait when polling for completion candidates."
+  :type 'float
+  :safe 'floatp)
+
+(defcustom utop-capf-max-wait-time 0.1
+  "Maximum time to wait before giving up on completion."
+  :type 'float
+  :safe 'floatp)
 
 (defface utop-prompt
   '((((background dark)) (:foreground "Cyan1"))
     (((background light)) (:foreground "blue")))
-  "The face used to highlight the prompt."
-  :group 'utop)
+  "The face used to highlight the prompt.")
 
 (defface utop-stdout
   nil
-  "The face used to highlight messages comming from stdout."
-  :group 'utop)
+  "The face used to highlight messages coming from stdout.")
 
 (defface utop-stderr
   nil
-  "The face used to highlight messages comming from stderr."
-  :group 'utop)
+  "The face used to highlight messages coming from stderr.")
 
 (defface utop-frozen
   '((t (:bold t)))
@@ -118,8 +130,7 @@ This hook is only run if exiting actually kills the buffer."
 
 (defface utop-error
   '((t (:foreground "#ff4040" :bold t :underline t)))
-  "The face used to highlight errors in phrases."
-  :group 'utop)
+  "The face used to highlight errors in phrases.")
 
 ;; +-----------------------------------------------------------------+
 ;; | Constants                                                       |
@@ -135,54 +146,38 @@ This hook is only run if exiting actually kills the buffer."
 ;; | Variables                                                       |
 ;; +-----------------------------------------------------------------+
 
-(defvar utop-process nil
+(defvar-local utop-process nil
   "The Lisp-object for the utop sub-process")
 
-(defvar utop-mode-map
-  (let ((map (make-sparse-keymap)))
-    (define-key map [return] 'utop-eval-input-or-newline)
-    (define-key map [(control ?m)] 'utop-eval-input-or-newline)
-    (define-key map [(control ?j)] 'utop-eval-input-auto-end)
-    (define-key map [home] 'utop-bol)
-    (define-key map [(control ?a)] 'utop-bol)
-    (define-key map [(meta ?p)] 'utop-history-goto-prev)
-    (define-key map [(meta ?n)] 'utop-history-goto-next)
-    (define-key map (kbd "TAB") 'utop-complete)
-    (define-key map [(control ?c) (control ?c)] 'utop-interrupt)
-    (define-key map [(control ?c) (control ?i)] 'utop-interrupt)
-    (define-key map [(control ?c) (control ?k)] 'utop-kill)
-    (define-key map [(control ?c) (control ?g)] 'utop-exit)
-    (define-key map [(control ?c) (control ?s)] 'utop)
-    (define-key map [(control ?c) ?m] 'utop-copy-old-input)
-    map)
-  "The utop local keymap.")
-
-(defvar utop-prompt-min 0
+(defvar-local utop-prompt-min 0
   "The point at the beginning of the current prompt.")
 
-(defvar utop-prompt-max 0
+(defvar-local utop-prompt-max 0
   "The point at the end of the current prompt.")
 
-(defvar utop-input-prompt-max 0
+(defvar-local utop-input-prompt-max 0
   "The point at the end of the last input prompt.")
 
-(defvar utop-output ""
+(defvar-local utop-output ""
   "The output of the utop sub-process not yet processed.")
 
-(defvar utop-command-number 0
+(defvar-local utop-command-number 0
   "The number of the current command.")
 
-(defvar utop-completion nil
+(defvar-local utop-completion nil
   "Current completion.")
 
-(defvar utop-completion-prefixes nil
+(defvar-local utop-capf-completion-candidates nil
+  "Current completion when using capf.")
+
+(defvar-local utop-completion-prefixes nil
   "Prefixes for current completion.")
 
-(defvar utop-inhibit-check nil
+(defvar-local utop-inhibit-check nil
   "When set to a non-nil value, always insert text, even if it is
 before the end of prompt.")
 
-(defvar utop-state nil
+(defvar-local utop-state nil
   "State of utop. It is one of:
 
 - edit: the user is typing a command
@@ -191,66 +186,65 @@ before the end of prompt.")
 - wait: ocaml is evaluating a phrase
 - done: ocaml has died.")
 
-(defvar utop-complete-buffer nil
+(defvar-local utop-complete-buffer nil
   "The buffer that requested completion.")
 
-(defvar utop-initial-command nil
+(defvar-local utop-initial-command nil
   "Initial phrase to evaluate.")
 
-(defvar utop-initial-mode nil
+(defvar-local utop-initial-mode nil
   "Mode to evaluate utop-initial-command in (nil or :multi).")
 
-(defvar utop-phrase-terminator ";;"
+(defvar-local utop-phrase-terminator ";;"
   "The OCaml phrase terminator.")
 
-(defvar utop-pending-entry nil
+(defvar-local utop-pending-entry nil
   "History entry")
 
-(defvar utop-pending-position nil
+(defvar-local utop-pending-position nil
   "The position of the cursor in the phrase sent to OCaml (where
 to add the newline character if it is not accepted).")
 
-(make-variable-buffer-local
- (defvar utop-package-list nil
-   "List of packages to load when visiting OCaml buffer.
-Useful as file variable."))
+(defvar-local utop-package-list nil
+  "List of packages to load when visiting OCaml buffer.
+Useful as file variable.")
 
-(defvar utop-next-phrase-beginning 'utop-compat-next-phrase-beginning
+(defvar-local utop-next-phrase-beginning 'utop-compat-next-phrase-beginning
   "The function used to find the beginning of the next phrase.")
 
-(defvar utop-discover-phrase 'utop-compat-discover-phrase
+(defvar-local utop-discover-phrase 'utop-compat-discover-phrase
   "The function used to discover a phrase.
 It should return a triple (begin-pos, end-pos,
 end-pos-with-comments)." )
 
-(defvar utop-skip-after-eval-phrase t
+(defvar-local utop-skip-after-eval-phrase t
   "Whether to skip to next phrase after evaluation.
 
 Non-nil means skip to the end of the phrase after evaluation in the
 Caml toplevel")
 
-(defvar utop--complete-k (lambda (_candidates) '())
+(defvar-local utop--complete-k (lambda (_candidates) '())
   "continuation function to populate the candidates for the company
 backend")
 
-(defvar utop-protocol-version "0"
+(defvar-local utop-protocol-version "0"
   "detected version of utop protocol. 0 for unknown or version pre")
 
-(defvar utop--read-version nil
+(defvar-local utop--read-version nil
   "whether we've tried to detect the utop version")
 
-(defvar utop--company-loaded nil)
+(defvar-local utop--company-loaded nil)
 
 (defun utop--supports-company ()
   (and
-   ;; version< only works on version numbers 
+   ;; version< only works on version numbers
    (condition-case nil
        (version<= "1" utop-protocol-version)
      (error t))
    (featurep 'company)))
 
 ;; +-----------------------------------------------------------------+
-;; | Compability with different ocaml major modes                    |
+;; | Compatibility with different ocaml major modes                  |
 ;; +-----------------------------------------------------------------+
 
 (defun utop-compat-resolve (choices)
@@ -311,15 +305,20 @@ modes you need to set these variables:
          (inhibit-read-only t))
      (progn ,@actions)))
 
+(defun utop-buffer ()
+  "Return the utop buffer."
+  (get-buffer utop-buffer-name))
+
 (defun utop-send-string (str)
-  "Send a string to the utop process. This function can only be
-called in the utop buffer and while the state is not 'done."
+  "Send a string to the utop process.
+This function can only be called in the utop buffer and while the
+state is not `done'."
   (process-send-string utop-process str))
 
 (defun utop-send-command (str)
   "Send a command to utop. If utop is not running or has exited,
 it is started."
-  (let ((buf (get-buffer utop-buffer-name)))
+  (let ((buf (utop-buffer)))
     (unless buf
       (setq buf (save-excursion (utop))))
     (with-current-buffer buf
@@ -332,7 +331,7 @@ it is started."
 
 (defun utop-goto-point-max-all-windows ()
   "Move the point to the end of buffer in all utop windows."
-  (let ((buffer (get-buffer utop-buffer-name)))
+  (let ((buffer (utop-buffer)))
     (walk-windows
      (lambda (window)
        (when (eq (window-buffer window) buffer)
@@ -374,51 +373,6 @@ it is started."
       (utop-send-string (concat "data:" line "\n")))
     (utop-send-string "end:\n")))
 
-(defun utop-last-type ()
-  "Extract last inferred type from the uTop toplevel"
-  (with-current-buffer utop-buffer-name
-    (save-excursion
-      (goto-char utop-prompt-min)
-      (forward-line -1)
-      (let ((line (buffer-substring-no-properties
-                   (line-beginning-position) (line-end-position))))
-        (message line)))))
-
-;; Poor man's identifier at point
-(defun utop-ident-looking (backward)
-  "Find limits of an OCaml identifier"
-  (save-excursion
-    (skip-chars-forward " \n\t")
-    (let ((iterating t)
-          (start-pos (point))
-          end-pos)
-      (while iterating
-        (setq end-pos (point))
-        (if (not backward)
-            (progn
-              (right-word 1)
-              (setq iterating (not (looking-back "[ \t\r\n].*" start-pos))))
-          (progn
-            (left-word 1)
-            (setq iterating (not (save-excursion
-                                   (search-forward-regexp "[ \t\r\n].*" start-pos t)))))))
-      end-pos)))
-
-(defun utop-ident-at-point ()
-  "Identifier at point"
-  (let ((start-pos (utop-ident-looking t))
-        (end-pos (utop-ident-looking nil)))
-    (buffer-substring-no-properties start-pos end-pos)))
-
-;; Currently not working - the communication is asynchronous so how to
-;; make sure without implementing another state that the type
-;; information has been already printed?
-(defun utop-type-at-point ()
-  "Find type of an identifier at point from uTop"
-  (utop-eval-string (utop-ident-at-point)))
-;;  (utop-last-type)
-
-
 ;; +-----------------------------------------------------------------+
 ;; | Edition control                                                 |
 ;; +-----------------------------------------------------------------+
@@ -516,7 +470,7 @@ it is started."
         (add-text-properties (point-min) utop-prompt-max utop-non-editable-properties)))
     ;; If OCaml is executing a phrase, follow its output
     (when (eq utop-state 'wait)
-      (let ((buffer (get-buffer utop-buffer-name)))
+      (let ((buffer (utop-buffer)))
         (walk-windows
          (lambda (window)
            (when (eq (window-buffer window) buffer)
@@ -545,19 +499,17 @@ it is started."
 
 (defun utop-insert-phrase-terminator ()
   "Insert the phrase terminator at the end of buffer."
-  ;; Search the longest suffix of the input which is a prefix of the
-  ;; phrase terminator
-  (let* ((end (point-max)) (pos (max utop-prompt-max (- end (length utop-phrase-terminator)))))
-    (while (not (string-prefix-p (buffer-substring-no-properties pos end) utop-phrase-terminator))
-      (setq pos (1+ pos)))
-    ;; Insert only the missing part
-    (insert (substring utop-phrase-terminator (- end pos)))))
+  (re-search-forward ";*[ \t\n\r]*\\'")
+  (goto-char (match-beginning 0))
+  (unless (looking-at-p ";;")
+    (insert ";;")))
 
 (defun utop-process-line (line)
   "Process one line from the utop sub-process."
   ;; Extract the command and its argument
   (string-match "\\`\\([a-z-]*\\):\\(.*\\)\\'" line)
-  (let ((command (match-string 1 line)) (argument (match-string 2 line)))
+  (let ((command (match-string 1 line))
+        (argument (match-string 2 line)))
     (pcase command
       ;; Output on stdout
       ("stdout"
@@ -650,20 +602,19 @@ it is started."
       ("completion"
        (catch 'done
          (dolist (prefix utop-completion-prefixes)
-           (when (string-prefix-p prefix argument)
-             (push argument utop-completion)
-             (throw 'done t)))))
+           ;; We need to handle specially prefixes like "List.m" as
+           ;; the responses from utop don't include the module prefix.
+           (let* ((pos (string-match  "[^.]*$" prefix))
+                  (module-prefix (substring prefix 0 pos)))
+             (when (string-prefix-p (substring prefix pos) argument)
+               (push (concat module-prefix argument) utop-completion)
+               (throw 'done t))))))
       ;; End of completion
       ("completion-stop"
        (utop-set-state 'edit)
        (if (utop--supports-company)
            (funcall utop--complete-k (nreverse utop-completion))
-         (progn
-           (if (> (length utop-completion) 1)
-               (with-current-buffer utop-complete-buffer
-                 (with-output-to-temp-buffer "*Completions*"
-                   (display-completion-list (nreverse utop-completion))))
-             (minibuffer-hide-completions))))
+         (setq utop-capf-completion-candidates (nreverse utop-completion)))
        (setq utop-completion nil)))))
 
 (defun utop-process-output (_process output)
@@ -705,8 +656,8 @@ If ALLOW-INCOMPLETE is non-nil and the phrase is not terminated,
 then a newline character will be inserted and edition will
 continue.
 
-If AUTO-END is non-nill then ALLOW-INCOMPLETE is ignored and a
-phrase terminator (;; or ; if using revised syntax) will be
+If AUTO-END is non-nil then ALLOW-INCOMPLETE is ignored and the
+phrase terminator (;;) will be
 automatically inserted by utop.
 
 If ADD-TO-HISTORY is t then the input will be added to history."
@@ -760,10 +711,7 @@ If ADD-TO-HISTORY is t then the input will be added to history."
     ;; We are now waiting for completion
     (utop-set-state 'comp)
     ;; Send all lines to utop
-    (utop-send-string
-     (if (utop--supports-company)
-         "complete-company:\n"
-       "complete:\n"))
+    (utop-send-string "complete-company:\n")
     ;; Keep track of the prefixes, so we can avoid returning
     ;; completion which don't have a match.
     (setq utop-completion-prefixes lines)
@@ -772,9 +720,8 @@ If ADD-TO-HISTORY is t then the input will be added to history."
       (utop-send-string (concat "data:" line "\n")))
     (utop-send-string "end:\n")))
 
-(defun utop-complete ()
-  "Complete current input."
-  (interactive)
+(defun utop-complete-start ()
+  "Conditionally begins to request completion candidates from utop."
   ;; Complete only if the cursor is after the prompt
   (when (and (eq utop-state 'edit) (>= (point) utop-prompt-max))
     ;; Use this buffer
@@ -783,6 +730,30 @@ If ADD-TO-HISTORY is t then the input will be added to history."
     (utop-complete-input
      (buffer-substring-no-properties utop-prompt-max (point)))))
 
+(defun utop-completion-at-point ()
+  "Complete thing at point."
+  (setq utop-capf-completion-candidates nil)
+  (utop-complete-start)
+
+  (let ((elapsed-time 0))
+    (while (and (eq utop-state 'comp)
+                (> utop-capf-max-wait-time elapsed-time))
+      (sleep-for utop-capf-wait-interval)
+      (setq elapsed-time (+ elapsed-time utop-capf-wait-interval))))
+
+  (when (>= (length utop-capf-completion-candidates) 1)
+    (list
+     utop-prompt-max
+     (point)
+     utop-capf-completion-candidates)))
+
+(defun utop-complete ()
+  "Complete current input."
+  (interactive)
+  (if (utop--supports-company)
+      (utop-complete-start)
+    (completion-at-point)))
+
 ;; +-----------------------------------------------------------------+
 ;; | Eval                                                            |
 ;; +-----------------------------------------------------------------+
@@ -790,7 +761,7 @@ If ADD-TO-HISTORY is t then the input will be added to history."
 (defun utop-prepare-for-eval ()
   "Prepare utop for evaluation."
   (save-excursion
-    (let ((buf (get-buffer utop-buffer-name)))
+    (let ((buf (utop-buffer)))
       (cond
        (buf
         ;; Make the buffer appear
@@ -869,7 +840,7 @@ If ADD-TO-HISTORY is t then the input will be added to history."
            (input (buffer-substring-no-properties start end))
            (edit-buffer (current-buffer)))
       ;; Start utop if needed
-      (let ((utop-buffer (get-buffer utop-buffer-name)))
+      (let ((utop-buffer (utop-buffer)))
         (unless utop-buffer
           ;; The buffer does not exist, read arguments before creating
           ;; it so the user can cancel starting utop
@@ -894,10 +865,36 @@ If ADD-TO-HISTORY is t then the input will be added to history."
   "Go to the beginning of line or to the end of the prompt."
   (interactive)
   (with-current-buffer utop-buffer-name
-    (if (= (point-at-bol) utop-prompt-min)
+    (if (= (line-beginning-position) utop-prompt-min)
         (goto-char utop-prompt-max)
       (move-beginning-of-line 1))))
 
+;; +-----------------------------------------------------------------+
+;; | Switch to REPL                                                  |
+;; +-----------------------------------------------------------------+
+
+(defun utop-switch-to-repl (eob-p)
+  "Switch to the inferior utop process buffer.
+With prefix argument EOB-P, positions cursor at end of buffer."
+  (interactive "P")
+  (if (get-buffer-process utop-buffer-name)
+      (pop-to-buffer utop-buffer-name)
+    (call-interactively #'utop))
+  (when eob-p
+    (push-mark)
+    (goto-char (point-max))))
+
+(defun utop-switch-to-recent-buffer ()
+  "Switch to the most recently used `utop-minor-mode' buffer."
+  (interactive)
+  (let ((recent-utop-minor-mode-buffer (seq-find (lambda (buf)
+                                                          (with-current-buffer buf (bound-and-true-p utop-minor-mode)))
+                                                        (buffer-list))))
+    (if recent-utop-minor-mode-buffer
+        (pop-to-buffer recent-utop-minor-mode-buffer)
+      (message "utop: No recent OCaml buffer found."))))
+
+
 ;; +-----------------------------------------------------------------+
 ;; | Process control                                                 |
 ;; +-----------------------------------------------------------------+
@@ -917,7 +914,7 @@ If ADD-TO-HISTORY is t then the input will be added to history."
 (defun utop-exit (&optional exit-code)
   "Try to gracefully exit utop.
 
-EXIT-CODE is the exit code that shoud be returned by utop. It
+EXIT-CODE is the exit code that should be returned by utop. It
 defaults to 0."
   (interactive)
   (with-current-buffer utop-buffer-name
@@ -925,8 +922,8 @@ defaults to 0."
       (utop-send-string (format "exit:%d\n" (or exit-code 0))))))
 
 (defun utop-sentinel (_process _msg)
-  "Callback for process' state change."
-  (let ((buffer (get-buffer utop-buffer-name)))
+  "Callback for process's state change."
+  (let ((buffer (utop-buffer)))
     ;; Do nothing if the buffer does not exist anymore
     (when buffer
       (with-current-buffer utop-buffer-name
@@ -1042,41 +1039,9 @@ defaults to 0."
   "Perform actions defined by local variables"
   (utop-query-load-package-list))
 
-;; +-----------------------------------------------------------------+
-;; | Menu                                                            |
-;; +-----------------------------------------------------------------+
-
-(defun utop-is-running ()
-  (let ((buf (get-buffer utop-buffer-name)))
-    (when buf
-      (with-current-buffer buf
-        (and utop-process (eq (process-status utop-process) 'run))))))
-
-(defun utop-about ()
-  (interactive)
-  (describe-variable 'utop-license))
-
-(defun utop-help ()
-  (interactive)
-  (describe-function 'utop))
-
-(easy-menu-define
-  utop-menu utop-mode-map
-  "utop menu."
-  '("utop"
-    ["Start OCaml" utop t]
-    ["Interrupt OCaml" utop-interrupt :active (utop-is-running)]
-    ["Kill OCaml" utop-kill :active (utop-is-running)]
-    ["Exit utop gracefully" utop-exit :active (utop-is-running)]
-    ["Evaluate Phrase" utop-eval-input-auto-end :active (and (utop-is-running) (eq utop-state 'edit))]
-    "---"
-    ["Customize utop" (customize-group 'utop) t]
-    "---"
-    ["About" utop-about t]
-    ["Help" utop-help t]))
 
 ;; +-----------------------------------------------------------------+
-;; | The mode                                                        |
+;; | The minor mode                                                  |
 ;; +-----------------------------------------------------------------+
 
 (defun utop-arguments ()
@@ -1128,54 +1093,113 @@ defaults to 0."
     (utop-insert "\nRestarting...\n\n")
     (utop-start arguments)))
 
-(defun utop-setup-ocaml-buffer ()
-  "Deprecated"
-  (error "`utop-setup-ocaml-buffer' is deprecated, you need to
-replace it by `utop-minor-mode'.
-
-See https://github.com/diml/utop for configuration information."))
+(defun utop--state ()
+  "Retrieve the state of the utop buffer."
+  ;; we can either be in the utop buffer itself
+  (if (derived-mode-p 'utop-mode)
+      utop-state
+    ;; or in another buffer
+    (let ((buf (utop-buffer)))
+      (when buf
+        (with-current-buffer buf
+         utop-state)))))
 
 ;;;###autoload
 (define-minor-mode utop-minor-mode
   "Minor mode for utop."
   :lighter " utop"
   :keymap (let ((map (make-sparse-keymap)))
-            (define-key map (kbd "C-c C-s") 'utop)
-            (define-key map (kbd "C-x C-e") 'utop-eval-phrase)
-            (define-key map (kbd "C-x C-r") 'utop-eval-region)
-            (define-key map (kbd "C-c C-b") 'utop-eval-buffer)
-            (define-key map (kbd "C-c C-k") 'utop-kill)
+            (define-key map (kbd "C-c C-s") #'utop)
+            (define-key map (kbd "C-x C-e") #'utop-eval-phrase)
+            (define-key map (kbd "C-c C-e") #'utop-eval-phrase)
+            (define-key map (kbd "C-x C-r") #'utop-eval-region)
+            (define-key map (kbd "C-c C-b") #'utop-eval-buffer)
+            (define-key map (kbd "C-c C-k") #'utop-kill)
+            (define-key map (kbd "C-c C-z") #'utop-switch-to-repl)
+            (easy-menu-define
+              utop-minor-mode-menu map
+              "utop menu."
+              '("utop"
+                ["Start utop" utop t]
+                ["Switch to utop" utop-switch-to-repl t]
+                ["Interrupt utop" utop-interrupt :active (utop-running-p)]
+                ["Kill utop" utop-kill :active (utop-running-p)]
+                ["Exit utop gracefully" utop-exit :active (utop-running-p)]
+                "---"
+                ["Evaluate phrase" utop-eval-phrase :active (and (utop-running-p) (eq (utop--state) 'edit))]
+                ["Evaluate region" utop-eval-region :active (and (utop-running-p) (eq (utop--state) 'edit))]
+                ["Evaluate buffer" utop-eval-buffer :active (and (utop-running-p) (eq (utop--state) 'edit))]
+                "---"
+                ["Customize utop" (customize-group 'utop) t]
+                "---"
+                ["About" utop-about t]
+                ["Help" utop-help t]))
             map)
   ;; Load local file variables
   (add-hook 'hack-local-variables-hook 'utop-hack-local-variables))
 
 (defvar company-backends)
 
+;; +-----------------------------------------------------------------+
+;; | The major mode                                                  |
+;; +-----------------------------------------------------------------+
+
+(defun utop-running-p ()
+  (let ((buf (utop-buffer)))
+    (when buf
+      (with-current-buffer buf
+        (and utop-process (eq (process-status utop-process) 'run))))))
+
+(define-obsolete-function-alias 'utop-is-running 'utop-running-p "2.0")
+
+(defun utop-about ()
+  (interactive)
+  (describe-variable 'utop-license))
+
+(defun utop-help ()
+  (interactive)
+  (describe-function 'utop))
+
+(defvar utop-mode-map
+  (let ((map (make-sparse-keymap)))
+    (define-key map (kbd "RET")     #'utop-eval-input-or-newline)
+    (define-key map (kbd "C-m")     #'utop-eval-input-or-newline)
+    (define-key map (kbd "C-j")     #'utop-eval-input-auto-end)
+    (define-key map (kbd "<home>")  #'utop-bol)
+    (define-key map (kbd "C-a")     #'utop-bol)
+    (define-key map (kbd "M-p")     #'utop-history-goto-prev)
+    (define-key map (kbd "M-n")     #'utop-history-goto-next)
+    (define-key map (kbd "TAB")     #'utop-complete)
+    (define-key map (kbd "C-c C-c") #'utop-interrupt)
+    (define-key map (kbd "C-c C-i") #'utop-interrupt)
+    (define-key map (kbd "C-c C-k") #'utop-kill)
+    (define-key map (kbd "C-c C-g") #'utop-exit)
+    (define-key map (kbd "C-c C-s") #'utop)
+    (define-key map (kbd "C-c C-z") #'utop-switch-to-recent-buffer)
+    (define-key map (kbd "C-c m")   #'utop-copy-old-input)
+    (easy-menu-define
+      utop-menu map
+      "utop menu."
+      '("utop"
+        ["Start utop" utop t]
+        ["Interrupt utop" utop-interrupt :active (utop-running-p)]
+        ["Kill utop" utop-kill :active (utop-running-p)]
+        ["Exit utop gracefully" utop-exit :active (utop-running-p)]
+        "---"
+        ["Evaluate Phrase" utop-eval-input-auto-end :active (and (utop-running-p) (eq utop-state 'edit))]
+        ["Switch to OCaml source buffer" utop-switch-to-recent-buffer t]
+        "---"
+        ["Customize utop" (customize-group 'utop) t]
+        "---"
+        ["About" utop-about t]
+        ["Help" utop-help t]))
+    map)
+  "The utop local keymap.")
+
 ;;;###autoload
 (define-derived-mode utop-mode fundamental-mode "utop"
   "Set the buffer mode to utop."
 
-  ;; Local variables
-  (make-local-variable 'utop-process)
-  (make-local-variable 'utop-prompt-min)
-  (make-local-variable 'utop-prompt-max)
-  (make-local-variable 'utop-input-prompt-max)
-  (make-local-variable 'utop-last-prompt)
-  (make-local-variable 'utop-output)
-  (make-local-variable 'utop-command-number)
-  (make-local-variable 'utop-inhibit-check)
-  (make-local-variable 'utop-state)
-  (make-local-variable 'utop-complete-buffer)
-  (make-local-variable 'utop-initial-command)
-  (make-local-variable 'utop-initial-mode)
-  (make-local-variable 'utop-phrase-terminator)
-  (make-local-variable 'utop-pending-position)
-  (make-local-variable 'utop-pending-entry)
-  (make-local-variable 'utop-protocol-version)
-
-  (make-local-variable 'utop--complete-k)
-  (make-local-variable 'utop--read-version)
-
   ;; Set the hook to call before changing the buffer
   (add-hook 'before-change-functions 'utop-before-change nil t)
 
@@ -1189,8 +1213,11 @@ See https://github.com/diml/utop for configuration information."))
   (with-eval-after-load 'company
     (add-to-list 'company-backends #'utop-company-backend))
 
+  (add-hook 'completion-at-point-functions #'utop-completion-at-point nil 'local)
+
   ;; Start utop
   (utop-start (utop-arguments)))
+
 ;; +-----------------------------------------------------------------+
 ;; | Starting utop                                                   |
 ;; +-----------------------------------------------------------------+
@@ -1199,7 +1226,7 @@ See https://github.com/diml/utop for configuration information."))
 (defun utop ()
   "A universal toplevel for OCaml.
 
-url: https://forge.ocamlcore.org/projects/utop/
+url: https://github.com/ocaml-community/utop
 
 utop is a enhanced toplevel for OCaml with many features,
 including context sensitive completion.
@@ -1212,7 +1239,7 @@ To complete an identifier, simply press TAB.
 Special keys for utop:
 \\{utop-mode-map}"
   (interactive)
-  (let ((buf (get-buffer utop-buffer-name)))
+  (let ((buf (utop-buffer)))
     (cond
      (buf
       ;; Jump to the buffer
@@ -1238,8 +1265,8 @@ Special keys for utop:
 (declare-function company-begin-backend "ext:company")
 (declare-function company-grab-symbol-cons "ext:company")
 
-(defun utop-company-backend (command &optional _arg &rest ignored)
-  "company backend for utop completions"
+(defun utop-company-backend (command &optional _arg &rest _ignored)
+  "Perform an action determined by COMMAND."
   (interactive (list 'interactive))
   (pcase command
     ('interactive (company-begin-backend 'utop-company-backend))
